import keyboard
import time
from . import agilent
from . import menlo
from .utils import CheckPreviousData
import rpi_stepper_motors.driver as driver
from .driver import Driver

WSADKeys = ['w', 'a', 's', 'd', 'e', 'q', 'r', 'f']
MANUAL_HELP = 'First: Q - clockwise, A - counterclockwise \n Second: W - clockwise, S - counterclockwise \n Third: E - clockwise, D - counterclockwise \n Fourth: R - clockwise, F - counterclockwise.'
SWEEP_FILENAME = 'sweep_data.txt'


def looper(nsteps, forward_span):
    """Loop over all motors `forward_span` steps forward."""
    for i in range(forward_span):
        driver.drivers.move(0, nsteps, 'clockwise')
        for j in range(forward_span):
            driver.drivers.move(1, nsteps, 'clockwise')
            for k in range(forward_span):
                driver.drivers.move(2, nsteps, 'clockwise')
                for l in range(forward_span):
                    driver.drivers.move(3, nsteps, 'clockwise')
                    yield (i, j, k, l)


def get_diagnostics(what):
    menlo.ConnectToMenlo()
    agilent.ConnectToAgilent()
    vals = [0, 0, -1]              # DC, AC, BW
    if "menlo" in what:
        vals[0] = menlo.GetDC()
        vals[1] = menlo.GetAC()
    if "agilent" in  what:
        vals[2] = agilent.GetBandWidth()

    return vals


def is_modelocked(vals):
    if vals[1] > 1200 or (vals[2] < 1e6 and vals[2] > 0):
        return True
    else:
        return False


def Loop(step_divider, nsteps, forward_span, diagnostics="menlo",
         break_on_success=False):
    # TODO Jak radius nie jest podane, to poruszać się w całym zakresie.
    full_circle = Driver.steps_per_lap(step_divider)
    print('Number of steps in full circle:', full_circle)
    print('Number of steps per macrostep:', nsteps)
    print('Move each stepper motor +-macrosteps:', forward_span)
    driver.drivers.set_step_sizes(step_divider)
    print(break_on_success)
    for (i, j, k, l) in looper(nsteps, forward_span):
        diag_vals = get_diagnostics(diagnostics)
        add_to_file = []
        add_to_file.extend(driver.drivers.positions)
        add_to_file.extend(diag_vals)
        add_to_file = ' '.join([str(x) for x in add_to_file]) + '\n'
        print(add_to_file.strip())
        with open(SWEEP_FILENAME, 'a') as f:
            f.write(add_to_file)
        if break_on_success and is_modelocked(diag_vals):
            print(f'Modelocked with:', add_to_file.strip())
            return


def SearchingDataFromFile():
    print('Searching through data file. ')
    try: 
        combinations = CheckPreviousData('data.txt')
        for combination in combinations:
            print(combination)
            for i, pos in enumerate(combination):
                driver.drivers.move_to_position(i, int(pos))
            print(f'\nSet to: {combination}. Press [p] to quit the manual part.')
            event = keyboard.read_event()  
            if event.event_type == keyboard.KEY_DOWN:  
                userInput = event.name

            if userInput in ['w', 'a', 's', 'd', 'e', 'q', 'r', 'f', 'p']:
                WSAD(userInput)

            if userInput == 'esc':
                break

        print('All done.')

    except KeyboardInterrupt:
        print("\nExiting the script.")

    finally:
        driver.drivers.cleanup_pins()


def WSAD(event):
    if event == 'q':
        driver.drivers.move(0, 1, 'clockwise')
    elif event == 'a':
        driver.drivers.move(0, 1, 'counterclockwise')
    elif event == 'w':
        driver.drivers.move(1, 1, 'clockwise')
    elif event == 's':
        driver.drivers.move(1, 1, 'counterclockwise')
    elif event == 'e':
        driver.drivers.move(2, 1, 'clockwise')
    elif event == 'd':
        driver.drivers.move(2, 1, 'counterclockwise')
    elif event == 'r':
        driver.drivers.move(3, 1, 'clockwise')
    elif event == 'f':
        driver.drivers.move(3, 1, 'counterclockwise')
    elif event == 'p':
        return


def HalfManual(step_divider, nsteps, forward_span, diagnostics):
    print(MANUAL_HELP)
    print('Press [esc] to skip the manual control part. \n Press P to end the manual control.')
    driver.drivers.set_step_sizes(step_divider)
    try:
        while True:
            Loop(step_divider, nsteps, forward_span, diagnostics,
                 break_on_success=True)
            while True:
                event = keyboard.read_event()  
                if event.event_type == keyboard.KEY_DOWN:  
                    userInput = event.name
                    if userInput in WSADKeys:
                        WSAD(userInput)
                    if userInput == 'esc':
                        break
    except KeyboardInterrupt:
        print("\nExiting the script.")
    finally:
        driver.drivers.cleanup_pins()


def Manual(step_divider):
    print(MANUAL_HELP)
    print('[CTR + C] to exit.')   
    driver.drivers.set_step_sizes(step_divider)
    try:
        while True:
            event = keyboard.read_event()  
            if event.event_type == keyboard.KEY_DOWN:  
                userInput = event.name
                if userInput in WSADKeys:
                    WSAD(userInput)
                if userInput == 'esc':
                    break
    except KeyboardInterrupt:
        print("\nExiting the script.")
    finally:
        driver.drivers.cleanup_pins()


def Automat(step_divider, nsteps, forward_span, diagnostics):
    print('Saving data to file.')
    try:
        Loop(step_divider, nsteps, forward_span, diagnostics,
             break_on_success=False)
    except KeyboardInterrupt:
        print('\nExiting the script.')
    finally:
        driver.drivers.cleanup_pins()


def Random():
    print('Go to random spot.')
    try:
        for i in range(4):
            driver.drivers.move_to_random_position(i)
    finally:
        driver.drivers.cleanup_pins()
