"""Define Driver class for individual drivers and Drivers class for all drivers. All control of drivers should go through the drivers instance of Drivers class."""
import RPi.GPIO as GPIO
import time
import random

drivers = None


class Driver:
    DIVIDER_MAP = {
        256: (GPIO.HIGH, GPIO.HIGH, GPIO.LOW, GPIO.LOW),
        128: (GPIO.HIGH, GPIO.LOW, GPIO.LOW, GPIO.LOW),
        64: (GPIO.HIGH, GPIO.HIGH, GPIO.LOW, GPIO.HIGH),
        32: (GPIO.LOW, GPIO.HIGH, GPIO.LOW, GPIO.LOW),
        16: (GPIO.HIGH, GPIO.HIGH, GPIO.HIGH, GPIO.HIGH),
        8: (GPIO.HIGH, GPIO.LOW, GPIO.HIGH, GPIO.HIGH),
        4: (GPIO.LOW, GPIO.HIGH, GPIO.LOW, GPIO.HIGH),
        2: (GPIO.HIGH, GPIO.LOW, GPIO.HIGH, GPIO.LOW),
        1: (GPIO.LOW, GPIO.LOW, GPIO.LOW, GPIO.LOW)
    }

    def __init__(self, first, second, third, fourth, reset):
        self.pins = (first, second, third, fourth)
        self.reset_pin = reset
        self.position_counter = 0
        self.step_divider = 1
        self.SetPins()
        self.CleanUpPins()

    def SetPins(self):
        """Sets the Raspberry Pi pins as outputs.

        Parameters: None

        Returns: None.
        """
        GPIO.setmode(GPIO.BCM)
        for pin in self.pins:
            GPIO.setup(pin, GPIO.OUT)
        GPIO.setup(self.reset_pin, GPIO.OUT)

    def CleanUpPins(self):
        """Sets the Raspberry Pi pins to low.

        Parameters: None

        Returns: None
        """
        for pin in self.pins:
            GPIO.output(pin, GPIO.LOW)
        GPIO.output(self.reset_pin, GPIO.LOW)

    def SetStepSize(self, step_divider):
        """
        Parameters:
        - step_divider: step size divider for microstepping.

        Returns: None
        """
        try:
            for pin, state in zip(self.pins, self.DIVIDER_MAP[step_divider]):
                GPIO.output(pin, state)
            GPIO.output(self.reset_pin, GPIO.HIGH)
            time.sleep(0.1)
            GPIO.output(self.reset_pin, GPIO.LOW)
            GPIO.output(self.pins[2], GPIO.LOW)
            GPIO.output(self.pins[3], GPIO.LOW)
            self.step_divider = step_divider
        except KeyError:
            raise KeyError("Allowed step dividers: " +
                           str(self.DIVIDER_MAP.keys()))

    def SetDirection(self, direction):
        if direction == "clockwise":
            GPIO.output(self.pins[3], GPIO.HIGH)
        elif direction == "counterclockwise":
            GPIO.output(self.pins[3], GPIO.LOW)

    def increment_counter(self):
        self.position_counter = (self.position_counter + Driver.microsteps_per_step(self.step_divider)) % Driver.steps_per_lap(self.step_divider)

    def decrement_counter(self):
        self.position_counter = (self.position_counter - Driver.microsteps_per_step(self.step_divider)) % Driver.steps_per_lap(self.step_divider)

    def Move(self, nsteps, direction):
        self.SetDirection(direction)
        for _ in range(nsteps):
            GPIO.output(self.pins[2], GPIO.HIGH)
            time.sleep(0.001)
            GPIO.output(self.pins[2], GPIO.LOW)
            if direction == "clockwise":
                self.increment_counter()
            elif direction == "counterclockwise":
                self.decrement_counter()

    @staticmethod
    def steps_per_lap(step_divider):
        return 1704*step_divider

    @staticmethod
    def microsteps_per_step(step_divider):
        """Calculates how single given step influences the stepper motor's
        position: multiplies the given size by 256. The smallest single motor
        step corresponding to step divider 256 updates the position by 1; the
        biggest step updates by 256.

        Parameters:
        - step_divider: step size divider for microstepping

        Returns:
        - number of microsteps in single step.
        """
        return 256/step_divider

    def go_to_random_position(self):
        full_lap = Driver.steps_per_lap(256)
        self.GoToPosition(random.randint(0, full_lap))

    def GoToPosition(self, new_position):
        """new_position is given in units of smallest step

        accuracy is limited by current microstepping"""
        diff = new_position - self.position_counter
        direction = "clockwise" if diff >= 0 else "counterclockwise"
        diff = (abs(diff) * self.step_divider) // 256
        self.Move(diff, direction)


class Drivers:
    def __init__(self, positions_log):
        self.drivers = []
        self.drivers.append(Driver(26, 19, 13, 6))
        self.drivers.append(Driver(17, 4, 3, 2))
        self.drivers.append(Driver(1, 7, 12, 16))
        self.drivers.append(Driver(20, 21, 5, 0))

        self.positions_log = positions_log
        self.read_positions()

        for driver in self.drivers:
            driver.SetPins()

    @property
    def positions(self):
        return [driver.position_counter for driver in self.drivers]

    def read_positions(self):
        with open(self.positions_log, 'r') as f:
            positions = f.readlines()[-1].strip().split()
            positions = [int(position) for position in positions]
            for pos, driver in zip (positions, drivers):
                driver.position_counter = pos

            return positions

    def write_positions(self):
        with open(self.positions_log, 'a') as f:
            f.write(' '.join([str(driver.position_counter)
                              for driver in self.drivers]) +
                    '\n')

    def cleanup_pins(self):
        for driver in self.drivers:
            driver.CleanUpPins()
        GPIO.cleanup()

    def set_step_sizes(self, step_divider):
        for driver in self.drivers:
            driver.SetStepSize(step_divider)

    def move(self, i, nsteps, direction):
        self.drivers[i].Move(nsteps, direction)
        self.write_positions()

    def move_to_position(self, i, position):
        self.drivers[i].GoToPosition(position)
        self.write_positions()

    def move_to_random_position(self, i):
        self.drivers[i].go_to_random_position()
        self.write_positions()


def init_drivers():
    global drivers
    drivers = Drivers()
