import keyboard
import time
from datetime import datetime
from . import agilent
from . import menlo
from .utils import CheckPreviousData
import rpi_stepper_motors.driver as driver
from .driver import Driver

current_time = datetime.now()

WSADKeys = ['w', 'a', 's', 'd', 'e', 'q', 'r', 'f']
MANUAL_HELP = 'First: Q - clockwise, A - counterclockwise \n Second: W - clockwise, S - counterclockwise \n Third: E - clockwise, D - counterclockwise \n Fourth: R - clockwise, F - counterclockwise.'
SWEEP_FILENAME = 'sweep_data' + '_' + str(current_time.year) + '_' + str(current_time.month) + '_' + str(current_time.day) + '_' + str(current_time.hour) + '_' + str(current_time.minute) + '.txt'

def get_diagnostics(what):
    menlo.ConnectToMenlo()
    agilent.ConnectToAgilent()
    vals = [0, 0, -1]              # DC, AC, BW
    if "menlo" in what:
        vals[0] = menlo.GetDC()
        vals[1] = menlo.GetAC()
    if "agilent" in  what:
        vals[2] = agilent.GetBandWidth()

    return vals


def is_modelocked(vals):
    if vals[1] > 1200 or (vals[2] < 1e6 and vals[2] > 0):
        return True
    else:
        return False


def Loop(step_divider, nsteps, forward_span, diagnostics="menlo",
         break_on_success=False):
    # TODO Jak radius nie jest podane, to poruszać się w całym zakresie.
    full_circle = Driver.steps_per_lap(step_divider)
    print('Number of steps in full circle:', full_circle, flush=True)
    print('Number of steps per macrostep:', nsteps, flush=True)
    print('Move each stepper motor +-macrosteps:', forward_span, flush=True)
    driver.drivers.set_step_sizes(step_divider)
    print(break_on_success)
    for (i, j, k, l) in looper(nsteps, forward_span):
        diag_vals = get_diagnostics(diagnostics)
        if diag_vals[1] > 50:
            time.sleep(0.25)
        add_to_file = []
        add_to_file.extend(driver.drivers.positions)
        add_to_file.extend(diag_vals)
        add_to_file = ' '.join([str(x) for x in add_to_file]) + '\n'
        print(add_to_file.strip())
        with open(SWEEP_FILENAME, 'a') as f:
            f.write(add_to_file)
        if break_on_success and is_modelocked(diag_vals):
            print(f'Modelocked with:', add_to_file.strip())
            return

def WSAD(event, step_divider):
    if event == 'q':
        response = driver.drivers.move(0, step_divider,  1, 'clockwise')
    elif event == 'a':
        response = driver.drivers.move(0, step_divider, 1, 'counterclockwise')
    elif event == 'w':
        response = driver.drivers.move(1, step_divider, 1, 'clockwise')
    elif event == 's':
        response = driver.drivers.move(1, step_divider, 1, 'counterclockwise')
    elif event == 'e':
        response = driver.drivers.move(2, step_divider, 1, 'clockwise')
    elif event == 'd':
        response = driver.drivers.move(2, step_divider, 1, 'counterclockwise')
    elif event == 'r':
        response = driver.drivers.move(3, step_divider, 1, 'clockwise')
    elif event == 'f':
        response = driver.drivers.move(3, step_divider, 1, 'counterclockwise')
    elif event == 'p':
        return
    if response == 'end':
        print('You reached full period.')


def Manual(step_divider):
    print(MANUAL_HELP, flush=True)
    print('[CTR + C] to exit.', flush=True)   
    driver.drivers.set_step_sizes(step_divider)
    try:
        while True:
            event = keyboard.read_event()  
            if event.event_type == keyboard.KEY_DOWN:  
                userInput = event.name
                if userInput in WSADKeys:
                    WSAD(userInput, step_divider)
                if userInput == 'esc':
                    break
    except KeyboardInterrupt:
        print("\nExiting the script.")
    finally:
        driver.drivers.cleanup_pins()

def Random():
    print('Go to random spot.', flush=True)
    try:
        for i in range(4):
            driver.drivers.move_to_random_position(i)
    finally:
        driver.drivers.cleanup_pins()


def GoTo(step_divider, combination):
    print(f"Go to set spot.", flush=True)
    driver.drivers.set_step_sizes(step_divider)
    try:
        for i, pos in enumerate(combination):
            driver.drivers.move_to_position(i, int(pos))
        print(f'\nSet to: {combination}.')
    finally:
        driver.drivers.cleanup_pins()


def MenloLooper(step_divider, step_widths):
    directions = ['clockwise', 'clockwise', 'clockwise', 'clockwise']
    while True:
        for i in range(len(directions)):
            response = driver.drivers.move(i, step_divider, step_widths[i], directions[i])
            if (response == 'end' and directions[i] == 'clockwise'):
                directions[i] = 'counterclockwise'
            elif (response == 'end' and directions[i] == 'counterclockwise'):
                directions[i] = 'clockwise'
            yield i

def MenloLoop(step_divider, diagnostics = 'menlo', break_on_success=False):
    period = 853*step_divider
    step_widths = [2, 7, 5, 3]
    print('Steps in full period: ', period, flush=True)
    print('Step widths for each driver: ', step_widths, flush=True)
    driver.drivers.set_step_sizes(step_divider)
    for i in MenloLooper(step_divider, step_widths):
        diag_vals = get_diagnostics(diagnostics)
        if diag_vals[1] > 50:
            time.sleep(0.25)
        add_to_file = []
        add_to_file.extend(driver.drivers.positions)
        add_to_file.extend(diag_vals)
        add_to_file = ' '.join([str(x) for x in add_to_file]) + '\n'
        print(add_to_file.strip())
        with open(SWEEP_FILENAME, 'w') as f:
            f.write(add_to_file)
        if break_on_success and is_modelocked(diag_vals):
            print(f'Modelocked with:', add_to_file.strip())
            return
        

def MenloHalfManual(step_divider, diagnostics):
    print(MANUAL_HELP, flush=True)
    print('Press [esc] to skip the manual control part. \n Press P to end the manual control.', flush=True)
    try:
        while True:
            MenloLoop(step_divider, diagnostics, break_on_success=True)
            while True:
                event = keyboard.read_event()  
                if event.event_type == keyboard.KEY_DOWN:  
                    userInput = event.name
                    if userInput in WSADKeys:
                        WSAD(userInput, step_divider)
                    if userInput == 'esc':
                        break
    except KeyboardInterrupt:
        print("\nExiting the script.")
    finally:
        driver.drivers.cleanup_pins()



